/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef ASSEMBLYJOBGRAPH_CPP
#define ASSEMBLYJOBGRAPH_CPP

#include "AssemblyJobGraph.h"
#include "AssemblyJobNullCopy.h"

#include "AssemblyException.h"
#include <queue>
#include <iostream>
using namespace::std;


AssemblyJob* AssemblyJobGraph::AssemblyJobGraphFactory(AssemblyGraphNode* beginNode, AssemblyGraphNode* endNode,
						       AssemblyJobFactory* jobFactory){
  AssemblyJob* aj = new AssemblyJobGraph(beginNode, endNode, jobFactory);
  set<ScoredSeq*> input;
  aj->allInputSeqs( &input );
  if (input.size() < 2){
    delete aj;
    aj = new AssemblyJobNullCopy( &input );
  }
  return aj;
}

AssemblyJobGraph::AssemblyJobGraph(){}

AssemblyJobGraph::AssemblyJobGraph(AssemblyGraphNode* beginNode, AssemblyGraphNode* endNode,
				   AssemblyJobFactory* jobFactory){
  constructorHelper(beginNode,endNode,jobFactory);
}


void AssemblyJobGraph::constructorHelper(AssemblyGraphNode* beginNode, AssemblyGraphNode* endNode, AssemblyJobFactory* ajf){
  _inputRemoved = false;
  _wasRun = false;
  _jobFactory = new AssemblyJobFactory(ajf);
  _isShallow = false;

  // check this requirement - if it is not met, bad things will happen
  if (_jobFactory->strandedness() != AssemblyJob::singleStranded){
    throw AssemblyException::ArgError("AJGraph, factory cannot be ds");
  }

  // figure out how many nodes there are
  _numNodes = 1;
  AssemblyGraphNode* markNode = beginNode;
  while (markNode != endNode){
    markNode = markNode->getPlusLink();
    _numNodes++;
  }

  // make an input set for each edge (including tips) - include just the nested seqs, delete flips
  _edgeSets = new set<ScoredSeq*>[_numNodes + 1];
  _minOverlaps = new long[_numNodes];
  _needsToBeAssembled = new bool[_numNodes + 1];
  for (long n = 0; n <= _numNodes; ++n){ _needsToBeAssembled[n] = true; }

  // FIRST, organize the input seqs (delete the flips)
  _rawInputSetsPlus = new set<ScoredSeq*>[_numNodes + 1];
  _rawInputSetsMinus = new set<ScoredSeq*>[_numNodes + 1];

  // take care of the leading edge
  set<ScoredSeqFlip*> localCollector;
  beginNode->getMinusEdgeSeqs(&localCollector);
  for (set<ScoredSeqFlip*>::iterator it = localCollector.begin(); it != localCollector.end(); ++it){
    switch( (*it)->getSense() ){
    case '+': _rawInputSetsPlus[0].insert( (*it)->getNested() ); break;
    case '-': _rawInputSetsMinus[0].insert( (*it)->getNested() ); break;
    default: throw AssemblyException::ArgError("AJG, bad sense char");
    }
    delete *it;
  }
  _rawInputTotalPlus.insert( _rawInputSetsPlus[0].begin(), _rawInputSetsPlus[0].end() );
  _rawInputTotalMinus.insert( _rawInputSetsMinus[0].begin(), _rawInputSetsMinus[0].end() );


  // take care of the linking edges
  AssemblyGraphNode* firstNode = beginNode;
  _minOverlaps[0] = firstNode->getCenterContig()->size();
  long setIndex = 1;
  while (firstNode != endNode){
    AssemblyGraphNode* secondNode = firstNode->getPlusLink();
    _minOverlaps[setIndex] = secondNode->getCenterContig()->size();
    set<ScoredSeqFlip*> localCollector;
    firstNode->getPlusEdgeSeqs(&localCollector);
    secondNode->getMinusEdgeSeqs(&localCollector);
    for (set<ScoredSeqFlip*>::iterator it = localCollector.begin(); it != localCollector.end(); ++it){
      switch( (*it)->getSense() ){
      case '+': _rawInputSetsPlus[setIndex].insert( (*it)->getNested() ); break;
      case '-': _rawInputSetsMinus[setIndex].insert( (*it)->getNested() ); break;
      default: throw AssemblyException::ArgError("AJG, bad sense char");
      }
      delete *it;
    }
    _rawInputTotalPlus.insert( _rawInputSetsPlus[setIndex].begin(), _rawInputSetsPlus[setIndex].end() );
    _rawInputTotalMinus.insert( _rawInputSetsMinus[setIndex].begin(), _rawInputSetsMinus[setIndex].end() );
    setIndex++;
    firstNode = secondNode;
  }

  // take care of the trailing edge
  localCollector.clear();
  endNode->getPlusEdgeSeqs(&localCollector);
  for (set<ScoredSeqFlip*>::iterator it = localCollector.begin(); it != localCollector.end(); ++it){
    switch( (*it)->getSense() ){
    case '+': _rawInputSetsPlus[_numNodes].insert( (*it)->getNested() ); break;
    case '-': _rawInputSetsMinus[_numNodes].insert( (*it)->getNested() ); break;
    default: throw AssemblyException::ArgError("AJG, bad sense char");
    }
    delete *it;
  }
  _rawInputTotalPlus.insert( _rawInputSetsPlus[_numNodes].begin(), _rawInputSetsPlus[_numNodes].end() );
  _rawInputTotalMinus.insert( _rawInputSetsMinus[_numNodes].begin(), _rawInputSetsMinus[_numNodes].end() );
}


AssemblyJobGraph::~AssemblyJobGraph(){
  delete _jobFactory;

  delete [] _edgeSets;
  delete [] _minOverlaps;
  delete [] _needsToBeAssembled;
  delete [] _rawInputSetsPlus;
  delete [] _rawInputSetsMinus;
}



void AssemblyJobGraph::prepareShallow(AssemblyJob::AssemblyThreadedness threadedness){ 
  if (! _isShallow){
    // figure out what sequences are used redundantly and how redundantly across all of the jobs.
    // this section should not be threaded because seqToCount is being modified
    map<ScoredSeq*,long> seqToCount;
    for (long edgeBin = 0; edgeBin <= _numNodes; ++edgeBin){
      for (set<ScoredSeq*>::iterator it = _rawInputSetsPlus[edgeBin].begin(); it != _rawInputSetsPlus[edgeBin].end(); ++it){
	map<ScoredSeq*,long>::iterator foundSeq = seqToCount.find( *it );
	if (foundSeq == seqToCount.end()){
	  seqToCount.insert(pair<ScoredSeq*,long>(*it,long(1)));
	} else { foundSeq->second++; }
      }
      for (set<ScoredSeq*>::iterator it = _rawInputSetsMinus[edgeBin].begin(); it != _rawInputSetsMinus[edgeBin].end(); ++it){
	map<ScoredSeq*,long>::iterator foundSeq = seqToCount.find( *it );
	if (foundSeq == seqToCount.end()){
	  seqToCount.insert(pair<ScoredSeq*,long>(*it,long(1)));
	} else { foundSeq->second++; }
      }
    }
    // speed up the search time in the next step by removing elements with a count of just 1
    ScoredSeq** toBeRemovedFromMap = new ScoredSeq*[ seqToCount.size() ];
    long tbrIndex = 0;
    for (map<ScoredSeq*,long>::iterator it = seqToCount.begin(); it != seqToCount.end(); ++it){
      if (it->second == long(1)){
	toBeRemovedFromMap[tbrIndex] = it->first;
	tbrIndex++;
      }
    }
    for (long n = 0; n < tbrIndex; ++n){ seqToCount.erase( toBeRemovedFromMap[n] ); }
    delete [] toBeRemovedFromMap;

    // i am going to replace every sequence with its normalized equivalent in every set and just be done with it
    // this section can be threaded since seqToCount is just being read from
    if (threadedness == AssemblyJob::notThreaded){
      for (long edgeBin = 0; edgeBin <= _numNodes; ++edgeBin){
	prepareShallowHelper(edgeBin, &seqToCount);
      }
    } else {
      #pragma omp parallel for schedule(dynamic)
      for (long edgeBin = 0; edgeBin <= _numNodes; ++edgeBin){
	prepareShallowHelper(edgeBin, &seqToCount);
      }
    }
    _isShallow = true;
  }
}



void AssemblyJobGraph::prepareShallowHelper(long edgeBin, map<ScoredSeq*,long>* seqToCount){ 
  for (set<ScoredSeq*>::iterator it = _rawInputSetsPlus[edgeBin].begin(); it != _rawInputSetsPlus[edgeBin].end(); ++it){
    map<ScoredSeq*,long>::iterator foundSeq = seqToCount->find( *it );
    if (foundSeq == seqToCount->end()){ _edgeSets[edgeBin].insert( (*it)->shallowCopy() ); }
    else {
      ScoredSeqNormalized* tempNormalized = new ScoredSeqNormalized( *it );
      tempNormalized->addCounts( foundSeq->second );
      _edgeSets[edgeBin].insert( tempNormalized->shallowCopy() );
      delete tempNormalized;
    }
  }
  for (set<ScoredSeq*>::iterator it = _rawInputSetsMinus[edgeBin].begin(); it != _rawInputSetsMinus[edgeBin].end(); ++it){
    map<ScoredSeq*,long>::iterator foundSeq = seqToCount->find( *it );
    if (foundSeq == seqToCount->end()){ _edgeSets[edgeBin].insert( (*it)->flipCopy() ); }
    else {
      ScoredSeqNormalized* tempNormalized = new ScoredSeqNormalized( *it );
      tempNormalized->addCounts( foundSeq->second );
      _edgeSets[edgeBin].insert( tempNormalized->flipCopy() );
      delete tempNormalized;
    }
  }
}




void AssemblyJobGraph::makeShallow(bool shallowDelete){
  if (_wasRun){
    throw AssemblyException::CallingError("AJGraph::makeShallow can't be called after the job was run.");
  }
  prepareShallow(AssemblyJob::notThreaded);
}



void AssemblyJobGraph::runCenterAssemblyJob(set<ScoredSeq*>* seqSetToModify, long minOverlap, AssemblyJob::AssemblyThreadedness threadedness){

  // raise the bar on the min overlap requirement and nullify any max overlap
  ParamsMinOverlap* oldPmo = _jobFactory->getParamsMinOverlap();
  ParamsMinOverlap* newPmo = new ParamsMinOverlap(oldPmo,  minOverlap);
  AlignmentScoreMatrix* asMatrix = _jobFactory->getScoreMatrix();
  ParamsMinFractId* pmf = _jobFactory->getParamsMinFractId();
  ParamsDeBruijn* pdb = _jobFactory->getParamsDeBruijn();
  //ParamsDeBruijn* pdb = new ParamsDeBruijn(0);

  AssemblyJobFactory* metaFactory = new AssemblyJobFactory(newPmo, pmf, asMatrix, pdb, AssemblyJob::singleStranded, _jobFactory->getListener());
  delete oldPmo;
  delete newPmo;
  delete asMatrix;
  delete pmf;
  delete pdb;

  // make and run the job and collect output into the same set that was passed in as an arg
  AssemblyJob* ajC = metaFactory->assemblyJob( seqSetToModify );
  delete metaFactory;
  ajC->runJob(threadedness);
  set<ScoredSeq*> leftoverSeqs;
  ajC->discardedInputSeqs(&leftoverSeqs);
  for (set<ScoredSeq*>::iterator it = leftoverSeqs.begin(); it != leftoverSeqs.end(); ++it){ delete *it; }
  seqSetToModify->clear();
  ajC->allAssembledSeqs( seqSetToModify );
  delete ajC;
}



void AssemblyJobGraph::collapseOneNode(long tipIndex, long edgeIndex, AssemblyJob::AssemblyThreadedness threadedness){

  // make sure that the two input sets have been dealt with
  if ( _needsToBeAssembled[tipIndex] ){
    AssemblyJob* aj = _jobFactory->assemblyJob( &_edgeSets[tipIndex] );
    aj->runJob(threadedness);
    set<ScoredSeq*> leftoverSeqs;
    aj->discardedInputSeqs( &leftoverSeqs );
    for (set<ScoredSeq*>::iterator it = leftoverSeqs.begin(); it != leftoverSeqs.end(); ++it){ delete *it; }
    _edgeSets[tipIndex].clear();
    aj->allAssembledSeqs( &_edgeSets[tipIndex] );
    delete aj;
    _needsToBeAssembled[tipIndex] = false;
  }

  if ( _needsToBeAssembled[edgeIndex] ){
    AssemblyJob* aj = _jobFactory->assemblyJob( &_edgeSets[edgeIndex] );
    aj->runJob(threadedness);
    set<ScoredSeq*> leftoverSeqs;
    aj->discardedInputSeqs( &leftoverSeqs );
    for (set<ScoredSeq*>::iterator it = leftoverSeqs.begin(); it != leftoverSeqs.end(); ++it){ delete *it; }
    _edgeSets[edgeIndex].clear();
    aj->allAssembledSeqs( &_edgeSets[edgeIndex] );
    delete aj;
    _needsToBeAssembled[edgeIndex] = false;
  }

  // I want to retain just those seqs that derive in some way from the edgeIndex
  // set for the next tip set.  So I will make a collection for mapping before I
  // empty that set.  These are copies so I can delete the originals during assembly.
  set<ScoredSeq*> anchorSet;
  for(set<ScoredSeq*>::iterator it = _edgeSets[edgeIndex].begin(); it != _edgeSets[edgeIndex].end(); ++it){
    anchorSet.insert( (*it)->shallowCopy() );
  }
  // collapse the anchor set - this won't take long and will save mapping time
  AssemblyJob* anchorJob = _jobFactory->redundancyJob(&anchorSet);
  anchorJob->runJob(threadedness);
  anchorSet.clear();
  anchorJob->discardedInputSeqs(&anchorSet);
  for (set<ScoredSeq*>::iterator it = anchorSet.begin(); it != anchorSet.end(); ++it){ (*it)->deepDelete(); }
  anchorSet.clear();
  anchorJob->allAssembledSeqs(&anchorSet);
  delete anchorJob;

  set<ScoredSeq*> combineBeginInput;
  combineBeginInput.insert( _edgeSets[tipIndex].begin(), _edgeSets[tipIndex].end() );
  _edgeSets[tipIndex].clear();
  combineBeginInput.insert( _edgeSets[edgeIndex].begin(), _edgeSets[edgeIndex].end() );
  _edgeSets[edgeIndex].clear();

  // switch out the min overlap requirement and combine the two edge results, to generate
  // a new tip edge set of seqs.
  long overlapIndex;
  if (tipIndex > edgeIndex){ overlapIndex = edgeIndex; }
  else { overlapIndex = tipIndex; }

  runCenterAssemblyJob(&combineBeginInput, _minOverlaps[overlapIndex], threadedness);

  // use the properties of the assembled set to determine the alignment parameters;
  // min overlap will be specified for each search based on the query length
  ParamsMinFractId* pmf = _jobFactory->getParamsMinFractId();
  ParamsMinOverlap* pmo = _jobFactory->getParamsMinOverlap();

  // ungapped alignment to maximize speed
  /*
  ScoredSeqCollectionBwt* anchorSeqCollection = new ScoredSeqCollectionBwt( &anchorSet,
									    pmf->calculateMinFractId(&combineBeginInput),
									    pmo->calculateMinOverlap(combineBeginInput.size()) );
  */

  // gapped alignment to maximize retention
  AlignmentScoreMatrix* asMatrix = _jobFactory->getScoreMatrix();
  DynamicProgrammingAligner* dpa = new DynamicProgrammingAligner(pmf->calculateMinFractId(&combineBeginInput),
								 pmo->calculateMinOverlap(combineBeginInput.size()),
								 asMatrix);
  delete asMatrix;
  ScoredSeqCollectionBwt* anchorSeqCollection = new ScoredSeqCollectionBwt(&anchorSet, dpa);
  delete dpa;

  delete pmf;
  delete pmo;

  // set up for threaded searching (same structure whether threaded or not)
  ScoredSeq** cbiArray = new ScoredSeq*[ combineBeginInput.size() + 1 ];
  long cbiIndex = 0;
  for (set<ScoredSeq*>::iterator seqIt = combineBeginInput.begin(); seqIt != combineBeginInput.end(); ++seqIt){
    cbiArray[cbiIndex] = *seqIt;
    cbiIndex++;
  }
  bool* cbiKeepBool = new bool[ combineBeginInput.size() + 1 ];

  // perform the searches to find out which sequences have matches (results stored in the array; may be threaded)
  if (threadedness == AssemblyJob::threaded){
    #pragma omp parallel for schedule(dynamic)
    for (long n = 0; n < combineBeginInput.size(); ++n){
      set<Alignment*> alSet;
      anchorSeqCollection->getBestMatches(&alSet, cbiArray[n]);
      cbiKeepBool[n] = (alSet.size() > 0);
      for (set<Alignment*>::iterator alIt = alSet.begin(); alIt != alSet.end(); ++alIt){ delete *alIt; }
    }
  } else {
    for (long n = 0; n < combineBeginInput.size(); ++n){
      set<Alignment*> alSet;
      anchorSeqCollection->getBestMatches(&alSet, cbiArray[n]);
      cbiKeepBool[n] = (alSet.size() > 0);
      for (set<Alignment*>::iterator alIt = alSet.begin(); alIt != alSet.end(); ++alIt){ delete *alIt; }
    }
  }

  // place the contents according to the results (modifies the collections in shared memory; not threaded)
  for (long n = 0; n < combineBeginInput.size(); ++n){
    if ( cbiKeepBool[n] ){
      _edgeSets[edgeIndex].insert( cbiArray[n] );
      combineBeginInput.erase( cbiArray[n] );
    }
  }

  delete [] cbiArray;
  delete [] cbiKeepBool;
  for (set<ScoredSeq*>::iterator it = anchorSet.begin(); it != anchorSet.end(); ++it){ (*it)->deepDelete(); }
  delete anchorSeqCollection;

  _fullProductSet.insert( combineBeginInput.begin(), combineBeginInput.end() );
}



void AssemblyJobGraph::helpSortAssembledSeqs(set<ScoredSeq*>* seqsToKeep, set<ScoredSeq*>* seqsThatAreDone,
					     ScoredSeqCollectionBwt* anchorSeqCollection,
					     AssemblyJob::AssemblyThreadedness threadedness){
  /*
  // don't bother making alignments to sequences that were already found
  MatchSeqTest* matchSeqTest;
  if (threadedness == AssemblyJob::threaded){
    #pragma omp critical (AJGmodifyFoundSeqs)
    { matchSeqTest = new MatchSeqTestNotInSet(seqsToKeep); }
  } else {
    matchSeqTest = new MatchSeqTestNotInSet(seqsToKeep);
  }

  // do the search
  set<Alignment*> foundAls;
  novelSeqCollection->getMatches(&foundAls, anchorSeq, '+', matchSeqTest, anchorSeq->size(), ScoredSeqCollectionBwt::softMinOvl);
  delete matchSeqTest;
  set<ScoredSeq*> foundSeqs;
  for (set<Alignment*>::iterator it = foundAls.begin(); it != foundAls.end(); ++it){
    ScoredSeq* foundSeq = (*it)->seqB();
    foundSeqs.insert( foundSeq );
    delete *it;
  }
  
  // sort the sequences returned by the search
  if (threadedness == AssemblyJob::threaded){
    #pragma omp critical (AJGmodifyFoundSeqs)
    { seqsToKeep->insert(foundSeqs.begin(), foundSeqs.end()); }
  } else {
    seqsToKeep->insert(foundSeqs.begin(), foundSeqs.end());
  }
  */
}

// THE OLD ONE - DELETE AFTER FINISHING THE NEW ONE
void AssemblyJobGraph::helpSortAssembledSeqs(set<ScoredSeq*>* seqsToKeep, set<ScoredSeq*>* seqsThatAreDone, ScoredSeq* anchorSeq,
					     ScoredSeqCollectionBwt* novelSeqCollection,
					     AssemblyJob::AssemblyThreadedness threadedness){
  // don't bother making alignments to sequences that were already found
  MatchSeqTest* matchSeqTest;
  if (threadedness == AssemblyJob::threaded){
    #pragma omp critical (AJGmodifyFoundSeqs)
    { matchSeqTest = new MatchSeqTestNotInSet(seqsToKeep); }
  } else {
    matchSeqTest = new MatchSeqTestNotInSet(seqsToKeep);
  }

  // do the search
  set<Alignment*> foundAls;
  novelSeqCollection->getMatches(&foundAls, anchorSeq, '+', matchSeqTest, anchorSeq->size(), ScoredSeqCollectionBwt::softMinOvl);
  delete matchSeqTest;
  set<ScoredSeq*> foundSeqs;
  for (set<Alignment*>::iterator it = foundAls.begin(); it != foundAls.end(); ++it){
    ScoredSeq* foundSeq = (*it)->seqB();
    foundSeqs.insert( foundSeq );
    delete *it;
  }
  
  // sort the sequences returned by the search
  if (threadedness == AssemblyJob::threaded){
    #pragma omp critical (AJGmodifyFoundSeqs)
    { seqsToKeep->insert(foundSeqs.begin(), foundSeqs.end()); }
  } else {
    seqsToKeep->insert(foundSeqs.begin(), foundSeqs.end());
  }
}


void AssemblyJobGraph::runJob(AssemblyJob::AssemblyThreadedness threadedness){
  if (! _wasRun){
    if (! _isShallow){ prepareShallow(threadedness); }

    // approach the center job from the two tips moving inwards
    long beginEdgeIndex = 0;
    long endEdgeIndex = _numNodes;
    while (beginEdgeIndex != endEdgeIndex){
      collapseOneNode(beginEdgeIndex, beginEdgeIndex + 1, threadedness);
      beginEdgeIndex++;
      // re-check to make sure that the forward step didn't already collapse things
      if (beginEdgeIndex != endEdgeIndex){
	collapseOneNode(endEdgeIndex, endEdgeIndex - 1, threadedness);
	endEdgeIndex--;
      }
    }

    // the assembly has finished, and the products are distributed in two places.  the ones
    // that passed the filter by matching sequences from the adjacent input bin are still in
    // _edgeSets[endEdgeIndex], aka _edgeSets[beginEdgeIndex].  the ones that were generated
    // but cast aside because they didn't match the targets are in _fullRedundantProductSet.
    // they should be combined and made non-redundant before being released.
    _fullProductSet.insert( _edgeSets[beginEdgeIndex].begin(), _edgeSets[beginEdgeIndex].end() );
    _edgeSets[beginEdgeIndex].clear();

    // make the finished set non-redundant
    AssemblyJob* ajNr = _jobFactory->redundancyJob( &_fullProductSet );
    ajNr->runJob(threadedness);
    set<ScoredSeq*> discardedSeqs;
    ajNr->discardedInputSeqs( &discardedSeqs );
    for (set<ScoredSeq*>::iterator it = discardedSeqs.begin(); it != discardedSeqs.end(); ++it){ delete *it; }
    _fullProductSet.clear();
    ajNr->allAssembledSeqs( &_fullProductSet );
    delete ajNr;

    _wasRun = true;
  }
}

bool AssemblyJobGraph::wasRun(){ return _wasRun; }

void AssemblyJobGraph::allInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _rawInputTotalPlus.begin(), _rawInputTotalPlus.end() );
  outsideSet->insert( _rawInputTotalMinus.begin(), _rawInputTotalMinus.end() );
}
void AssemblyJobGraph::remainingInputSeqs( set<ScoredSeq*>* outsideSet ){
}
void AssemblyJobGraph::discardedInputSeqs( set<ScoredSeq*>* outsideSet ){
  allInputSeqs(outsideSet);
}
void AssemblyJobGraph::newlyAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _fullProductSet.begin(), _fullProductSet.end() );
}
void AssemblyJobGraph::allAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _fullProductSet.begin(), _fullProductSet.end() );
}



void AssemblyJobGraph::OK(){}



#endif
