/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

/* A carrier of useful functions for assembly.
 *
 * This is a static class (just a carrier for functions)
 */


#ifndef ASSEMBLYJOBHELPERSTRATEGIES_H
#define ASSEMBLYJOBHELPERSTRATEGIES_H

#include <set>
#include "ScoredSeq.h"
#include "Alignment.h"
using namespace::std;

class AssemblyJobHelperStrategies{

 public:

  /* same as above, but for just one alignment */
  static ScoredSeq* combine(Alignment* alignment);
  // seqA will be divided by denominator for normalization
  static ScoredSeq* combine(Alignment* alignment, int denominator);


  // deals with sub-string words
  // IN PROGRESS!!!!!!!!!
  // return values????
  void getSubtextWordQueriesHelper(ScoredSeq* querySeq, long maxMismatch);
  void getSubtextWordQueriesHelper(ScoredSeq* querySeq, long maxMismatch, long blockStartPos, long blockEndPos);




 private:

  static float _SCORE_MAX;

  // this is just a vehicle for carrying around data; its data fields are public
  class CombineDataCarrier{
  public:
    CombineDataCarrier();
    CombineDataCarrier(long maxSize);
    ~CombineDataCarrier();
    //CombineDataCarrier(ScoredSeq * preferredSeq);
    void addNuc(char nuc, float score);
    void addNuc(char nuc, float score, float link);
    void addLink(float link);
    void decrementLink();
    ScoredSeq* getSeq();
  private:
    long _size;
    char* _seq;
    float* _scores;
    float* _links;
    long _linkOffset;
    long _maxSize;
    //ScoredSeq * preferredSeq;
  };

  // this is a way of providing altenative methods for calculating scores
  class ScoreCalculator {
  public:
    virtual ~ScoreCalculator();
    virtual float adjustA(float a) = 0;
    virtual float adjustB(float b) = 0;
    virtual float aPlusB(float a, float b) = 0;
    virtual float aMinusB(float a, float b) = 0;
    virtual float bMinusA(float b, float a) = 0;
  };
  class ScoreCalcAdd : public ScoreCalculator {
  public:
    ScoreCalcAdd();
    ~ScoreCalcAdd();
    float adjustA(float a);
    float adjustB(float b);
    float aPlusB(float a, float b);
    float aMinusB(float a, float b);
    float bMinusA(float b, float a);
  };
  class ScoreCalcNormalizeA : public ScoreCalculator {
  public:
    ScoreCalcNormalizeA();
    ScoreCalcNormalizeA(int denominator); // divides seqA by denominator
    ~ScoreCalcNormalizeA();
    float adjustA(float a);
    float adjustB(float b);
    float aPlusB(float a, float b);
    float aMinusB(float a, float b);
    float bMinusA(float b, float a);
  private:
    int _denominator;
  };
  class ScoreCalcXy { // allows easy swapping of A vs B identity without ambiguous labelling
  public:
    ScoreCalcXy();
    ScoreCalcXy(ScoreCalculator* calc, bool xIsA);
    ~ScoreCalcXy();
    float adjustX(float x);
    float adjustY(float y);
    float xPlusY(float x, float y);
    float xMinusY(float x, float y);
    float yMinusX(float y, float x);
  private:
    bool _xIsA;
    ScoreCalculator* _calc;
  };


  static bool isPositionFinal(ScoredSeq * seq, long pos);
  static bool isPositionFinal(ScoredSeq * seqX, long posX, ScoredSeq * seqY, long posY);

  static ScoredSeq* combine(Alignment* alignment, ScoreCalculator* calc);
  static void combineGapBlock(Alignment* alignment,
		       long aStartPos, long aEndPos,
		       long bStartPos, long bEndPos,
		       CombineDataCarrier* carrier, ScoreCalculator* outsideCalc);
  static void getSegmentsFromBlock(map<long,float>* links, long startPos, long endPos, char sense,
			    ScoredSeq* seq, Alignment* alignment, ScoreCalcXy* calc);
  static void fillInGaps(CombineDataCarrier* carrier,
		  ScoredSeq* seq, long startPos, long endPos, char sense,
		  float penaltyConstant, map<long,float>* links, ScoreCalcXy* calc);

};


#endif
