/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef EXTENDCYCLE_CPP
#define EXTENDCYCLE_CPP

#include "ExtendCycle.h"
#include "ExtendJobCreator.h"
#include "Read.h"
#include "ScoredSeqSubseq.h"
#include "ScoredSeqNested.h"
#include "ScoredSeqMonoScore.h"
#include <typeinfo>
#include <sstream>
#include <omp.h>
#include <limits.h>
using namespace::std;

int ExtendCycle::_plusIndex = 0;
int ExtendCycle::_minusIndex = 1;
int ExtendCycle::_frontIndex = 1;
int ExtendCycle::_backIndex = 0;

int ExtendCycle::plusIndex(){ return _plusIndex; }
int ExtendCycle::minusIndex(){ return _minusIndex; }
int ExtendCycle::frontIndex(){ return _frontIndex; }
int ExtendCycle::backIndex(){ return _backIndex; }

ExtendCycle::ExtendCycle(){}

void ExtendCycle::constructorHelper(set<ParameterizedReadFile*>* readFileSet,
				    set<ScoredSeq*>* inputContigs,
				    set<ScoredSeq*>* doNotExtend){

  _cycleHasRun = false;

  _numReadFiles = readFileSet->size();
  _readFileArray = new ParameterizedReadFile*[ _numReadFiles ];
  _hitReadsByFile = new set<ScoredSeqWithCarriers*>*[ _numReadFiles ];
  for (int n = 0; n < _numReadFiles; ++n){ _hitReadsByFile[n] = new set<ScoredSeqWithCarriers*>; }

  int rfCounter = 0;
  for (set<ParameterizedReadFile*>::iterator rfIt = readFileSet->begin(); rfIt != readFileSet->end(); ++rfIt){
    _readFileArray[rfCounter] = (*rfIt);
    rfCounter++;
  }
  // make shallow copies of the input contigs so that they can be deleted
  for (set<ScoredSeq*>::iterator icIt = inputContigs->begin(); icIt != inputContigs->end(); ++icIt){
    ScoredSeqNormalized* inSeq = new ScoredSeqNormalized( (*icIt)->shallowCopy() );
    _inputContigs.insert( inSeq );
    if ( doNotExtend->count( (*icIt) ) != 0 ){ _doNotExtendContigs.insert( inSeq ); }
  }

  _numThreads = omp_get_max_threads();
}

ExtendCycle::ExtendCycle(set<ParameterizedReadFile*>* readFileSet, ReadPairFilter* rpf, set<ScoredSeq*>* inputContigs, AssemblerListener* listener) :
  _listener(listener) {
  _rpf = rpf->copy();
  set<ScoredSeq*> emptyDoNotExtend;
  constructorHelper(readFileSet, inputContigs, &emptyDoNotExtend);
}

ExtendCycle::ExtendCycle(set<ParameterizedReadFile*>* readFileSet, ReadPairFilter* rpf, set<ScoredSeq*>* inputContigs, set<ScoredSeq*>* doNotExtend, AssemblerListener* listener) :
  _listener(listener) {
  _rpf = rpf->copy();
  constructorHelper(readFileSet, inputContigs, doNotExtend);
}

ExtendCycle::~ExtendCycle(){
  delete _rpf;
  delete [] _readFileArray;
  for (int n = 0; n < _numReadFiles; ++n){ delete _hitReadsByFile[n]; }
  delete [] _hitReadsByFile;

  // get rid of the reads that were in use and make sure that their paired ends were used up
  set<ScoredSeq*> toDeepDelete;
  toDeepDelete.insert( _readsUsedThisCycle.begin(), _readsUsedThisCycle.end() );
  for (set<ScoredSeqNormalized*>::iterator it = _readsUsedThisCycle.begin(); it != _readsUsedThisCycle.end(); ++it){
    if ( (*it)->hasPairedEnd() ){ toDeepDelete.insert( (*it)->getPairedEnd() ); }
  }
  for (set<ScoredSeq*>::iterator readIt = toDeepDelete.begin(); readIt != toDeepDelete.end(); ++readIt){
    (*readIt)->deepDelete();
  }

  // get rid of the wrappers of the contigs (the map set outermost wrappers)
  for (set<ScoredSeqWithCarriers*>::iterator inIt = _mapSets.begin(); inIt != _mapSets.end(); ++inIt){
    delete (*inIt); // shallow delete since inner contigs were already deleted
  }
}

void ExtendCycle::runCycle(ParamsAlignment * paMini,
			   ParamsAlignment * paMeta,
			   ParamsDeBruijn* pdb,
			   long maxContigLinkages,
			   EcoFilter* preMetaFilter){
  // set up the carriers that will hold the read-contig mapping linkages
  set<ScoredSeqWithCarriers*> dneWithCarriers;
  for (set<ScoredSeqNormalized*>::iterator it = _inputContigs.begin(); it != _inputContigs.end(); ++it){
    ScoredSeqNormalized* contig = (*it); // the contig will be normalized
    ScoredSeqWithCarriers* carrier = new ScoredSeqWithCarriers(contig, 2); // 0->frontSet, 1->backSet
    _mapSets.insert( carrier );
    if ( _doNotExtendContigs.count(contig) != 0){ dneWithCarriers.insert(carrier); }
  }

  MapReadSetupHelper* mrsh;
  ParameterizedReadFile** nullRfa = new ParameterizedReadFile*[1];
  nullRfa[0] = NULL;
  if (dneWithCarriers.size() == _inputContigs.size()){
    mrsh = new MrsMainHelper(nullRfa, 0, _rpf, &dneWithCarriers);
  } else {
    mrsh = new MrsMainHelper(_readFileArray, _numReadFiles, _rpf, &dneWithCarriers);
  }
  mapReads(mrsh);
  delete mrsh;

  mrsh = new MrsAddendumHelper(_readFileArray, _numReadFiles, _hitReadsByFile);
  mapReads(mrsh);
  delete mrsh;

  runMiniAssemblies(paMini, pdb, maxContigLinkages);
  runMetaAssembly(paMeta, pdb, preMetaFilter);
  delete [] nullRfa;
  _cycleHasRun = true;
}



ExtendCycle::MapReadSetupHelper::~MapReadSetupHelper(){}
ScoredSeq** ExtendCycle::MapReadSetupHelper::getFileTriggers(ParameterizedReadFile** readFileArray, int numReadFiles){
  ScoredSeq** fileTriggers = new ScoredSeq*[ numReadFiles ];
  for (int n = 0; n < numReadFiles; ++n){
    readFileArray[n]->open(false);
    if ( readFileArray[n]->hasRead() ){
      fileTriggers[n] = readFileArray[n]->getRead();
    } else { fileTriggers[n] = 0; }
    readFileArray[n]->close();
  }
  return fileTriggers;
}
void ExtendCycle::MapReadSetupHelper::triggerFilesHelper(ScoredSeq** triggerArray, int numReadFiles){
  for (int n = 0; n < numReadFiles; ++n){
    string s = triggerArray[n]->getSeq('+');
    triggerArray[n]->deepUnbuffer();
  }
}



ExtendCycle::MrsMainHelper::MrsMainHelper(ParameterizedReadFile** readFileArray, int numReadFiles, ReadPairFilter* rpf, set<ScoredSeqWithCarriers*>* doNotExtendContigs) :
  _numReadFiles(numReadFiles) {
  _rpf = rpf->copy();
  // figure out the total file sizes for stdout update and set up for threaded file reading
  _numActiveFiles = 0;
  _totalFileSize = 0;
  _doNotExtendContigs.insert(doNotExtendContigs->begin(), doNotExtendContigs->end());

  _readFileArray = new ParameterizedReadFile*[ _numReadFiles + 1 ];
  for (int n = 0; n < _numReadFiles; ++n){ _readFileArray[n] = readFileArray[n]; }
  _fileIsActive = new bool[ _numReadFiles ]; // keeps track of how many files are still being read
  _fileToPmap = new ParamsMapping*[ _numReadFiles ];
  _fileTrigger = getFileTriggers(_readFileArray, _numReadFiles);

  #pragma omp parallel for schedule(dynamic)
  for (int n = 0; n < _numReadFiles; ++n){
    long numReads = _readFileArray[n]->numReads();
    #pragma omp critical (ECmapReads)
    {
      _totalFileSize += numReads;
      // pre-open the files that have reads
      if (numReads == 0){
	_fileIsActive[n] = false;
	_fileToPmap[n] = 0;
      } else {
	_fileToPmap[n] = _readFileArray[n]->getMapParams();
	_readFileArray[n]->openNormWithCarriers(2,false);
	_fileIsActive[n] = true;
	_numActiveFiles++;
      }
    }
  }
}
ExtendCycle::MrsMainHelper::~MrsMainHelper(){
  delete _rpf;
  for (int n = 0; n < _numReadFiles; ++n){ 
    _readFileArray[n]->close();
    delete _fileToPmap[n];
    delete _fileTrigger[n];
  }
  delete [] _fileTrigger;
  delete [] _readFileArray;
  delete [] _fileIsActive;
  delete [] _fileToPmap;
}

int ExtendCycle::MrsMainHelper::numFilesTotal(){ return _numReadFiles; }
int ExtendCycle::MrsMainHelper::numFilesActive(){ return _numActiveFiles; }
bool ExtendCycle::MrsMainHelper::isFileActive(int fileNum){ return _fileIsActive[fileNum]; }
//void ExtendCycle::MrsMainHelper::triggerFiles(){ triggerFilesHelper(_fileTrigger, _numReadFiles); }
void ExtendCycle::MrsMainHelper::triggerFiles(){
  for (int n = 0; n < _numReadFiles; ++n){
    ReadFileCommunicator::bufferFill(_readFileArray[n], ReadFile::bufferThreaded);
  }
}


long ExtendCycle::MrsMainHelper::totalReadCount(){ return _totalFileSize; }
ExtendJobMapper* ExtendCycle::MrsMainHelper::getNewEjm(set<ScoredSeqWithCarriers*>* parallelMapSet, int fileNum){
  return new ExtendJobMapper( parallelMapSet, _fileToPmap[fileNum]->fractId(), _readFileArray[fileNum]->ampliconSize() );
}

ScoredSeqWithCarriers** ExtendCycle::MrsMainHelper::getNewReadArray(int fileNum){
  ParameterizedReadFile* readFile = _readFileArray[fileNum];
  long arrayTargetSize = ( _fileToPmap[fileNum]->stepSize() / long(2) ) * long(2);
  ScoredSeqWithCarriers** seqsToMapArray = new ScoredSeqWithCarriers*[ arrayTargetSize + 1 ];
  // set up the mapper object
  long readCounter = 0;
  while ( readCounter < arrayTargetSize and readFile->hasRead() ){
    ScoredSeqWithCarriers* newRead1 = dynamic_cast<ScoredSeqWithCarriers*>( readFile->getRead() );
    // only paired-end reads will be mapped, if the file has a mixture
    if ( newRead1->hasPairedEnd() ){
      ScoredSeqWithCarriers* newRead2 = dynamic_cast<ScoredSeqWithCarriers*>( newRead1->getPairedEnd() );
      newRead1->bottomBuffer();
      newRead2->bottomBuffer();
      if ( _rpf->isPairOk(newRead1, newRead2) ){
	seqsToMapArray[ readCounter ] = newRead1;
	readCounter += 1;
	seqsToMapArray[ readCounter ] = newRead2;
	readCounter += 1;
      } else {
	newRead1->deepDelete();
	newRead2->deepDelete();
      }
    } else { newRead1->deepDelete(); }
  }
  // end the array with a null element
  seqsToMapArray[readCounter] = 0;
  return seqsToMapArray;
}

void ExtendCycle::MrsMainHelper::updateActiveFiles(){
  _numActiveFiles = 0;
  for (int n = 0; n < _numReadFiles; ++n){
    if (_fileIsActive[n]){
      if ( _readFileArray[n]->hasRead() ){ _numActiveFiles++; }
      else { _fileIsActive[n] = false; }
    }
  }
}
bool ExtendCycle::MrsMainHelper::includeContigCarrier(ScoredSeqWithCarriers* carrier){ return _doNotExtendContigs.count(carrier) == 0; }


void ExtendCycle::mapReads(MapReadSetupHelper* mrsh){

  // no need to do anything if there are no active files
  if (mrsh->numFilesActive() == 0){
    _listener->beginMapping(0);
    _listener->endMapping();
  } else {

    // set up for mapping - one EJM for each file, for each thread
    ExtendJobMapper** fileToEjm = new ExtendJobMapper*[ _numReadFiles ];
    #pragma omp parallel for schedule(dynamic)
    for (int fileN = 0; fileN < mrsh->numFilesTotal(); ++fileN){
      if ( mrsh->isFileActive(fileN) ){
	fileToEjm[ fileN ] = mrsh->getNewEjm(&_mapSets, fileN);
      } else { fileToEjm[ fileN ] = 0; }
    }
    _listener->beginMapping(mrsh->totalReadCount());

    while (mrsh->numFilesActive() > 0){
      // prepare a subset of reads for mapping - fill these in below
      ScoredSeqWithCarriers*** seqsToMapArray = new ScoredSeqWithCarriers**[ _numThreads ];
      long seqsToMapCount[_numThreads];
      long seqsWithHitsCount[_numThreads];

      // get the files to use this round, the max num of files allowed by the num threads
      set<int> thisRoundFileNums;
      int fn = 0;
      while (fn < mrsh->numFilesTotal() and thisRoundFileNums.size() < _numThreads){
	if ( mrsh->isFileActive(fn) ){ thisRoundFileNums.insert(fn); }
	++fn;
      }

      // allocate threads to each file
      int numRoundFiles = thisRoundFileNums.size();
      int numThreadsUsed = 0;
      int threadToFileNum[_numThreads];

      // set up an array of targets
      ExtendJobMapper* ejmArray[_numThreads ];

      while (numThreadsUsed < _numThreads and thisRoundFileNums.size() > 0){
        // get the files ready for read extraction by creating local thread arrays
	int numLocalThreads = thisRoundFileNums.size();
	if (numLocalThreads + numThreadsUsed > _numThreads){ numLocalThreads = _numThreads - numThreadsUsed; }
	int localIndex = 0;
	for (set<int>::iterator it = thisRoundFileNums.begin(); it != thisRoundFileNums.end(); ++it){
	  if (localIndex < numLocalThreads){
	    threadToFileNum[localIndex + numThreadsUsed] = *it;
	    localIndex++;
	  }
	}

        // get the queries for each thread (threaded by file)
        #pragma omp parallel for schedule(dynamic)
	for (int localN = 0; localN < numLocalThreads; ++localN){
  	  // get all of the necessary local variables
	  int threadN = localN + numThreadsUsed;
	  int fileN = threadToFileNum[threadN];
	  ejmArray[threadN] = fileToEjm[fileN];
	  seqsToMapArray[threadN] = mrsh->getNewReadArray(fileN);
	  long readCounter = 0;
	  while (seqsToMapArray[threadN][readCounter] != 0){ readCounter++; }
          #pragma omp critical (updateCountArray)
	  { seqsToMapCount[ threadN ] = readCounter; }
	}

	numThreadsUsed += numLocalThreads;

        // get rid of the finished file nums
	mrsh->updateActiveFiles();
	set<int> fileNumsToRemove;
	for (set<int>::iterator it = thisRoundFileNums.begin(); it != thisRoundFileNums.end(); ++it){
	  if (! mrsh->isFileActive(*it)){ fileNumsToRemove.insert( *it ); }
	}
	for (set<int>::iterator it = fileNumsToRemove.begin(); it != fileNumsToRemove.end(); ++it){ thisRoundFileNums.erase( *it ); }
      }

      // trigger the buffering of sequences OUTSIDE the parallel block
      mrsh->triggerFiles();

      // do the mapping
      long readsBeingMapped = 0;
      for (int n = 0; n < numThreadsUsed; ++n){ readsBeingMapped += seqsToMapCount[n]; }

      #pragma omp parallel for schedule(dynamic)
      for (int threadN = 0; threadN < numThreadsUsed; ++threadN){
	long returnedCount = ejmArray[ threadN ]->mapQueries( seqsToMapArray[ threadN ], seqsToMapCount[ threadN ] );
        #pragma omp critical (updateReturnCountArray)
	{ seqsWithHitsCount[ threadN ] = returnedCount; }
	int fileN = threadToFileNum[ threadN ];
	for (long n = 0; n < returnedCount; ++n){ _hitReadsByFile[ fileN ]->insert( seqsToMapArray[ threadN ][n] ); }
      }
      _listener->updateMapping(readsBeingMapped);
      for (int n = 0; n < numThreadsUsed; ++n){ delete [] seqsToMapArray[n]; }

      // clean up the EJMapper objects for non-active files.  I will set them
      // to zero so that I can still call delete on them in later mapping rounds.
      mrsh->updateActiveFiles();
      for (int n = 0; n < mrsh->numFilesTotal(); ++n){
	if (! mrsh->isFileActive(n)){
	  delete fileToEjm[n];
	  fileToEjm[n] = 0;
	}
      }
      delete [] seqsToMapArray;
    }
    delete [] fileToEjm;

    _listener->endMapping();
  }
}






ExtendCycle::MrsAddendumHelper::MrsAddendumHelper(ParameterizedReadFile** readFileArray, int numReadFiles, set<ScoredSeqWithCarriers*>** hitReadsByFile) :
  _numReadFiles(numReadFiles) {
  _fileTrigger = getFileTriggers(readFileArray, _numReadFiles);
  _numActiveFiles = 0;
  _totalNumQueries = 0;

  _readFileArray = new ParameterizedReadFile*[ _numReadFiles + 1 ];
  for (int n = 0; n < _numReadFiles; ++n){ _readFileArray[n] = readFileArray[n]; }

  // make sets of the paired-ends of the mapped reads, THOSE will be mapped
  _pairQueries = new set<ScoredSeqWithCarriers*>*[ _numReadFiles ];
  for (int n = 0; n < _numReadFiles; ++n){
    _pairQueries[n] = new set<ScoredSeqWithCarriers*>;
    for (set<ScoredSeqWithCarriers*>::iterator it = hitReadsByFile[n]->begin(); it != hitReadsByFile[n]->end(); ++it){
      _pairQueries[n]->insert( dynamic_cast<ScoredSeqWithCarriers*>( (*it)->getPairedEnd() ) );
      // I also want to re-map the read to the entire dataset so that its repetitive nature (ability
      // to map to many different contigs) can be fairly evaluated
      _pairQueries[n]->insert( *it );
    }
  }

  // set up the iterators that will be used to allocate the reads
  _readItByFileNum = new set<ScoredSeqWithCarriers*>::iterator[ _numReadFiles ];
  _fileIsActive = new bool[ _numReadFiles ];
  _fileToPmap = new ParamsMapping*[ _numReadFiles ];
  for (int n = 0; n < _numReadFiles; ++n){
    _readItByFileNum[n] = _pairQueries[n]->begin();
    _totalNumQueries += _pairQueries[n]->size();
    _fileToPmap[n] = readFileArray[n]->getMapParams();
    if (_readItByFileNum[n] == _pairQueries[n]->end()){
      _fileIsActive[n] = false;
    } else {
      _fileIsActive[n] = true;
      _numActiveFiles++;
    }
  }
}
ExtendCycle::MrsAddendumHelper::~MrsAddendumHelper(){
  for (int n = 0; n < _numReadFiles; ++n){ 
    delete _fileTrigger[n];
    delete _pairQueries[n];
    delete _fileToPmap[n];
  }
  delete [] _fileTrigger;
  delete [] _pairQueries;
  delete [] _readItByFileNum;
  delete [] _fileIsActive;
  delete [] _fileToPmap;
  delete [] _readFileArray;
}
int ExtendCycle::MrsAddendumHelper::numFilesTotal(){ return _numReadFiles; }
int ExtendCycle::MrsAddendumHelper::numFilesActive(){ return _numActiveFiles; }
long ExtendCycle::MrsAddendumHelper::totalReadCount(){ return _totalNumQueries; }
bool ExtendCycle::MrsAddendumHelper::isFileActive(int fileNum){ return _fileIsActive[fileNum]; }
ExtendJobMapper* ExtendCycle::MrsAddendumHelper::getNewEjm(set<ScoredSeqWithCarriers*>* parallelMapSet, int fileNum){
  return new ExtendJobMapper( parallelMapSet, _fileToPmap[fileNum]->fractId() );
}
ScoredSeqWithCarriers** ExtendCycle::MrsAddendumHelper::getNewReadArray(int fileNum){
  ParamsMapping* pmap = _fileToPmap[ fileNum ];
  ScoredSeqWithCarriers** readArray = new ScoredSeqWithCarriers*[ pmap->stepSizeAddendum() + 1 ];
  long readCounter = 0;
  while ( readCounter < pmap->stepSizeAddendum() and _readItByFileNum[ fileNum ] != _pairQueries[ fileNum ]->end() ){
    ScoredSeqWithCarriers* newRead = *_readItByFileNum[ fileNum ];
    readArray[ readCounter ] = newRead;
    readCounter += 1;
    newRead->bottomBuffer();
    ++_readItByFileNum[ fileNum ];
  }
  readArray[readCounter] = 0;
  return readArray;
}
//void ExtendCycle::MrsAddendumHelper::triggerFiles(){ triggerFilesHelper(_fileTrigger, _numReadFiles); }
void ExtendCycle::MrsAddendumHelper::triggerFiles(){
  for (int n = 0; n < _numReadFiles; ++n){
    ReadFileCommunicator::bufferFill(_readFileArray[n], ReadFile::bufferThreaded);
  }
}


void ExtendCycle::MrsAddendumHelper::updateActiveFiles(){
  _numActiveFiles = 0;
  for (int fileN = 0; fileN < _numReadFiles; ++fileN){
    if (_fileIsActive[fileN]){
      if (_readItByFileNum[fileN] == _pairQueries[fileN]->end()){
	_fileIsActive[fileN] = false;
      } else { _numActiveFiles++; }
    }
  }
}
bool ExtendCycle::MrsAddendumHelper::includeContigCarrier(ScoredSeqWithCarriers* carrier){ return true; }



void ExtendCycle::runOneMiniAssembly(AssemblyJob* aj, set<ScoredSeq*>* metaInputByThread, AssemblyJob::AssemblyThreadedness threadedness){
  // the input params of this factory need to be further utilized
  int threadNum = omp_get_thread_num();

  ostringstream outMessage1;
  outMessage1 << "Thread " << threadNum << ": Job Start";
  _listener->verboseMessage(outMessage1.str());
  _listener->assemblyJobBeginStats(aj);

  // make shallow copies of everything so that they can be deleted without worry
  #pragma omp critical (ECsingleAssemblyHelper)
  { aj->makeShallow(false); } // <-that bool is shallowDeleteInput

  aj->runJob(threadedness);
  aj->allAssembledSeqs( metaInputByThread );

  //singleAssemblyHelper(aj, metaInputByThread, threadedness);

  _listener->assemblyJobEndStats(aj);
  ostringstream outMessage2;
  outMessage2 << "Thread " << threadNum << ": Job End";
  _listener->verboseMessage(outMessage2.str());
  _listener->updateEdgeAssembly(1);
}



void ExtendCycle::runMiniAssemblies(ParamsAlignment* pa, ParamsDeBruijn* pdb, long maxContigLinkages){

  AssemblyJobFactory * overhangFactory = new AssemblyJobFactory(pa, pdb, AssemblyJob::singleStranded, _listener); //single-stranded
  set<ScoredSeqWithCarriers*> garbageDump;
  set<AssemblyJob*>* jobSet = ExtendJobCreator::createJobs(&_mapSets, &_doNotExtendContigs,
							   overhangFactory, maxContigLinkages,
							   &garbageDump);
  delete overhangFactory;


  // make an initial unsorted version of the array
  AssemblyJob** jobArray = new AssemblyJob*[ jobSet->size() + 1 ];
  long totalNumJobs = 0;
  for (set<AssemblyJob*>::iterator jobIt = jobSet->begin(); jobIt != jobSet->end(); ++jobIt){
    jobArray[totalNumJobs] = *jobIt;
    totalNumJobs++;
  }
  delete jobSet;

  // i want the job array to be sorted by size, so first i need to know
  // how many jobs of each size there are.  while i'm at it, i am also
  // adding normalization counts to the SSNormalized job components.
  map<long, long> sizeToJobCount;
  long* arrayOfJobSizes = new long[ totalNumJobs ];
  long totalOfJobSizes = 0;
  for (long oc = 0; oc < totalNumJobs; ++oc){
    set<ScoredSeq*> inputSeqs;
    jobArray[oc]->allInputSeqs( &inputSeqs );

    for (set<ScoredSeq*>::iterator seqIt = inputSeqs.begin(); seqIt != inputSeqs.end(); ++seqIt){
      dynamic_cast<ScoredSeqNormalized*>( *seqIt )->addCount();
    }

    // DEFINE the job size as a long
    // i will sort by the product of the biggest contig times the summed lengths of all the rest
    // jobs with just one contig are null, and they will be appropriately sized as zero by this metric
    long jobSize = 0;
    if (inputSeqs.size() > 1){
      long totalContigSize = 0;
      long maxContigSize = 0;
      for (set<ScoredSeq*>::iterator seqIt = inputSeqs.begin(); seqIt != inputSeqs.end(); ++seqIt){
	long contigSize = (*seqIt)->size();
	totalContigSize += contigSize;
	if (contigSize > maxContigSize){ maxContigSize = contigSize; }
      }
      if (maxContigSize != 0){
	long upperBound = LONG_MAX / maxContigSize;
	long remainingLength = totalContigSize - maxContigSize;
	if (remainingLength < upperBound){
	  jobSize = maxContigSize * (totalContigSize - maxContigSize);
	} else { jobSize = LONG_MAX; }
      }
    }

    arrayOfJobSizes[oc] = jobSize;
    totalOfJobSizes += jobSize;
    map<long,long>::iterator stjcIt = sizeToJobCount.find( jobSize );
    if (stjcIt == sizeToJobCount.end()){ sizeToJobCount.insert( pair<long,long>(jobSize,1) ); }
    else { stjcIt->second++; }
  }


  // create another data structure to make sorting of the list easier
  map<long, long> sizeToStartIndex;
  long currentStartIndex = totalNumJobs;
  for (map<long,long>::iterator it = sizeToJobCount.begin(); it != sizeToJobCount.end(); ++it){
    currentStartIndex -= it->second;
    sizeToStartIndex.insert( pair<long,long>(it->first, currentStartIndex) );
  }

  // now i can use the information that i gathered above to sort the array
  AssemblyJob** sortedJobArray = new AssemblyJob*[ totalNumJobs ];
  long* arrayOfSortedJobSizes = new long[ totalNumJobs ];
  for (long oc = 0; oc < totalNumJobs; ++oc){
    map<long,long>::iterator indexIt = sizeToStartIndex.find( arrayOfJobSizes[oc] );
    arrayOfSortedJobSizes[ indexIt->second ] = arrayOfJobSizes[oc];
    sortedJobArray[ indexIt->second ] = jobArray[oc];
    indexIt->second++;
  }

  // replace the unsorted with the sorted versions
  delete [] jobArray;
  jobArray = sortedJobArray;
  delete [] arrayOfJobSizes;
  arrayOfJobSizes = arrayOfSortedJobSizes;



  // organize the jobs into those to be threaded and those to not be threaded
  long remainingJobSizeTotal = totalOfJobSizes;
  long startParallelIndex = 0;
  int numThreads = omp_get_max_threads();
  // if there is only one thread, no jobs will be run in parallel
  if (numThreads == 1){ startParallelIndex = totalNumJobs; }
  else if (totalNumJobs > 0){
    remainingJobSizeTotal -= arrayOfJobSizes[startParallelIndex];
    while (startParallelIndex < totalNumJobs and arrayOfJobSizes[startParallelIndex] * numThreads >= remainingJobSizeTotal){
      startParallelIndex++;
      if (startParallelIndex < totalNumJobs){ remainingJobSizeTotal -= arrayOfJobSizes[startParallelIndex]; }
    }
  }
  if (totalNumJobs - startParallelIndex < numThreads){ startParallelIndex = totalNumJobs; }

  _listener->beginEdgeAssembly(totalNumJobs);

  // an output collection for each thread; this will be consolidated after
  set<ScoredSeq*>* metaInputByThread[ _numThreads ];
  for (int tn = 0; tn < _numThreads; ++tn){ metaInputByThread[tn] = new set<ScoredSeq*>; }

  // first run the long jobs individually threaded
  for (long oc = 0; oc < startParallelIndex; ++oc){
    runOneMiniAssembly(jobArray[oc], metaInputByThread[0], AssemblyJob::threaded);
  }
  // now run the smaller jobs in parallel (biggest launched first)
  #pragma omp parallel for schedule(dynamic)
  for (long oc = startParallelIndex; oc < totalNumJobs; ++oc){
    int threadNum = omp_get_thread_num();
    runOneMiniAssembly(jobArray[oc], metaInputByThread[threadNum], AssemblyJob::notThreaded);
  }


  // clean up array out here
  for (long oc = 0; oc < totalNumJobs; ++oc){ delete jobArray[oc]; }
  delete [] jobArray;
  delete [] arrayOfJobSizes;

  // clean up the reads
  set<ScoredSeq*> deepDelete;
  for (set<ScoredSeqWithCarriers*>::iterator it = garbageDump.begin(); it != garbageDump.end(); ++it){
    deepDelete.insert( *it );
    if ( (*it)->hasPairedEnd() ){ deepDelete.insert( (*it)->getPairedEnd() ); }
  }
  for (set<ScoredSeq*>::iterator it = deepDelete.begin(); it != deepDelete.end(); ++it){ (*it)->deepDelete(); }

  // consolidate the output and clean up
  for (int n = 0; n < _numThreads; ++n){
    _metaInputContigs.insert( metaInputByThread[n]->begin(), metaInputByThread[n]->end() );
    delete metaInputByThread[n];
  }

  _listener->endEdgeAssembly();
}





void ExtendCycle::runMetaAssembly(ParamsAlignment* pa, ParamsDeBruijn* pdb, EcoFilter* preMetaFilter){

  // an output collection for each thread; this will be consolidated after
  int numThreads = 1;
  //int numThreads = omp_get_max_threads();

  // get rid of the too-$hort sequences
  set<ScoredSeq*> filteredMetaInput;
  set<ScoredSeq*> unsuitableMetaInput;
  preMetaFilter->filterSeqs(&_metaInputContigs, &filteredMetaInput, &unsuitableMetaInput, 0);
  for (set<ScoredSeq*>::iterator cit = unsuitableMetaInput.begin(); cit != unsuitableMetaInput.end(); ++cit){ (*cit)->deepDelete(); }
  _metaInputContigs.clear();

  _listener->beginMetaAssembly( filteredMetaInput.size() );
  AssemblyJobFactory * contigFactory = new AssemblyJobFactory(pa, pdb, AssemblyJob::doubleStranded, _listener);
  AssemblyJob* aj = contigFactory->redundancyJob( &filteredMetaInput );
  singleAssemblyHelper(aj, &_assembledContigs, AssemblyJob::threaded);
  delete aj;
  delete contigFactory;

  // delete all of the input contigs - this will include the _doNotExtendContigs
  for (set<ScoredSeqNormalized*>::iterator icIt = _inputContigs.begin(); icIt != _inputContigs.end(); ++icIt){
    (*icIt)->deepDelete();
  }
  _inputContigs.clear();

  _listener->endMetaAssembly( _assembledContigs.size() );

}



void ExtendCycle::singleAssemblyHelper(AssemblyJob* assemblyJob, set<ScoredSeq*>* outputRepository, AssemblyJob::AssemblyThreadedness threadedness){
  /*
  // make shallow copies of everything so that they can be deleted without worry
  #pragma omp critical (ECsingleAssemblyHelper)
  { assemblyJob->makeShallow(shallowDeleteInput); }
  */
  assemblyJob->runJob(threadedness);

  // get all the contigs (even the single-read ones)
  assemblyJob->allAssembledSeqs( outputRepository );

  // get the input data that was combined into the output contigs and delete it
  set<ScoredSeq*> discardedInput;
  assemblyJob->discardedInputSeqs( &discardedInput );
  for (set<ScoredSeq*>::iterator disIt = discardedInput.begin(); disIt != discardedInput.end(); ++disIt){
    (*disIt)->deepDelete();
  }
}




bool ExtendCycle::cycleHasRun(){ return _cycleHasRun; }
void ExtendCycle::getContigs(set<ScoredSeq*>* outputContigs){
  outputContigs->insert(_assembledContigs.begin(), _assembledContigs.end());
}


#endif
