/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef EXTENDCYCLE_H
#define EXTENDCYCLE_H

#include <set>
#include <map>
#include "ScoredSeq.h"
#include "ReadFile.h"
#include "ReadPairFilter.h"
#include "DynamicProgrammingAligner.h"
#include "ScoredSeqCollectionBwt.h"
#include "AssemblyJobFactory.h"
#include "AssemblyJob.h"
#include "ScoredSeqNormalized.h"
#include "ScoredSeqWithCarriers.h"
#include "ScoredSeqFlip.h"
#include "ParamsMapping.h"
#include "ParamsMinOverlap.h"
#include "ParamsMinFractId.h"
#include "ParamsAlignment.h"
#include "ParamsDeBruijn.h"
#include "ParameterizedReadFile.h"
#include "AssemblerListener.h"
#include "ExtendJobMapper.h"
#include "EcoFilter.h"
using namespace::std;
class ExtendJobMapper;

class ExtendCycle {

 public:
  ExtendCycle();
  ExtendCycle(set<ParameterizedReadFile*>* readFileSet, ReadPairFilter* rpf, set<ScoredSeq*>* inputContigs, AssemblerListener* listener);
  ExtendCycle(set<ParameterizedReadFile*>* readFileSet, ReadPairFilter* rpf, set<ScoredSeq*>* inputContigs, set<ScoredSeq*>* doNotExtend, AssemblerListener* listener);
  ~ExtendCycle();

  // make these accessible but not mutable
  static int plusIndex();
  static int minusIndex();
  static int frontIndex();
  static int backIndex();

  // in case there is nothing to filter

  void runCycle(ParamsAlignment * paMini,
		ParamsAlignment * paMeta,
		ParamsDeBruijn* pdb,
		long maxContigLinkages,
		EcoFilter* preMetaFilter);
  bool cycleHasRun();
  void getContigs(set<ScoredSeq*>* outputContigs);

 private:
  class MapReadSetupHelper;

  void mapReads(MapReadSetupHelper* mrsh);
  void runMiniAssemblies(ParamsAlignment* pa, ParamsDeBruijn* pdb, long maxContigLinkages);
  void runMetaAssembly(ParamsAlignment* pa, ParamsDeBruijn* pdb, EcoFilter* preMetaFilter);


  AssemblerListener* _listener;
  ReadPairFilter* _rpf;

  bool _cycleHasRun;
  void constructorHelper(set<ParameterizedReadFile*>* readFileSet,
			 set<ScoredSeq*>* inputContigs,
			 set<ScoredSeq*>* doNotExtend);

  set<ReadFile*> _readFiles; // needs to be provided initially

  set<ScoredSeqNormalized*> _inputContigs;
  set<ScoredSeqNormalized*> _doNotExtendContigs;
  ParameterizedReadFile** _readFileArray;
  set<ScoredSeqWithCarriers*>** _hitReadsByFile;
  int _numReadFiles;

  int _numThreads;

  // for general reference for the SSWC access to distinct collections
  static int _frontIndex; // 1
  static int _backIndex;  // 0
  static int _plusIndex;  // 0
  static int _minusIndex; // 1

  // for read mapping
  long _mappingStepSize;
  long _totalQueryNum;

  // WC encases Normalized encases Shallow
  //   carried: WC encasing Normalized encasing Read
  //              carried: WC encasing Normalized Shallow
  // 0->frontSet, 1->backSet
  set<ScoredSeqWithCarriers*> _mapSets;

  set<ScoredSeqNormalized*> _readsUsedThisCycle;

  // each Flip encases a Normalized encases a Read/Shallow
  set< set<ScoredSeqFlip*>* > _allOverhangJobs;

  set<ScoredSeq*> _metaInputContigs;
  set<ScoredSeq*> _assembledContigs;

  void runOneMiniAssembly(AssemblyJob* aj, set<ScoredSeq*>* metaInputByThread, AssemblyJob::AssemblyThreadedness threadedness);
  void singleAssemblyHelper(AssemblyJob* assemblyJob, set<ScoredSeq*>* outputRepository, AssemblyJob::AssemblyThreadedness threadedness);

  // used by mapReads to separate behavior between initial mapping and 
  class MapReadSetupHelper{
  public:
    virtual ~MapReadSetupHelper();
    virtual int numFilesTotal() = 0;
    virtual int numFilesActive() = 0;
    virtual long totalReadCount() = 0;
    virtual bool isFileActive(int fileNum) = 0;
    virtual ExtendJobMapper* getNewEjm(set<ScoredSeqWithCarriers*>* parallelMapSet, int fileNum) = 0;
    virtual ScoredSeqWithCarriers** getNewReadArray(int fileNum) = 0;
    virtual void triggerFiles() = 0;
    virtual void updateActiveFiles() = 0;
    virtual bool includeContigCarrier(ScoredSeqWithCarriers* carrier) = 0;
    static ScoredSeq** getFileTriggers(ParameterizedReadFile** readFileArray, int numReadFiles);
    static void triggerFilesHelper(ScoredSeq** triggerArray, int numReadFiles);
  };

  class MrsMainHelper : public MapReadSetupHelper {
  public:
    MrsMainHelper(ParameterizedReadFile** readFileArray, int numReadFiles, ReadPairFilter* rpf, set<ScoredSeqWithCarriers*>* doNotExtendContigs);
    ~MrsMainHelper();
    int numFilesTotal();
    int numFilesActive();
    long totalReadCount();
    bool isFileActive(int fileNum);
    ExtendJobMapper* getNewEjm(set<ScoredSeqWithCarriers*>* parallelMapSet, int fileNum);
    ScoredSeqWithCarriers** getNewReadArray(int fileNum);
    void triggerFiles();
    void updateActiveFiles();
    bool includeContigCarrier(ScoredSeqWithCarriers* carrier);
  private:
    ReadPairFilter* _rpf;
    int _numReadFiles;
    int _numActiveFiles;
    long _totalFileSize;
    bool* _fileIsActive;
    ParamsMapping** _fileToPmap;
    ParameterizedReadFile** _readFileArray;
    ScoredSeq** _fileTrigger;
    set<ScoredSeqWithCarriers*> _doNotExtendContigs;
  };

  class MrsAddendumHelper : public MapReadSetupHelper {
  public:
    MrsAddendumHelper(ParameterizedReadFile** readFileArray, int numReadFiles, set<ScoredSeqWithCarriers*>** hitReadsByFile);
    ~MrsAddendumHelper();
    int numFilesTotal();
    int numFilesActive();
    long totalReadCount();
    bool isFileActive(int fileNum);
    ExtendJobMapper* getNewEjm(set<ScoredSeqWithCarriers*>* parallelMapSet, int fileNum);
    ScoredSeqWithCarriers** getNewReadArray(int fileNum);
    void triggerFiles();
    void updateActiveFiles();
    bool includeContigCarrier(ScoredSeqWithCarriers* carrier);
  private:
    int _numReadFiles;
    int _numActiveFiles;
    long _totalNumQueries;
    bool* _fileIsActive;
    ScoredSeq** _fileTrigger;
    ParamsMapping** _fileToPmap;
    ParameterizedReadFile** _readFileArray;
    set<ScoredSeqWithCarriers*>** _pairQueries;
    set<ScoredSeqWithCarriers*>::iterator*  _readItByFileNum;
  };

};

#endif
