/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This is the text-based interface for the PRICE assembler.
 *
 */


#include "AssemblyException.h"
#include "AssemblyJob.h"
#include "ReadFile.h"
#include "ReadFileFastaSingle.h"
#include "ReadFileFastqSingle.h"
#include "ReadFilePriceqSingle.h"
#include "ReadFileDual.h"
#include "AssemblyJobFactory.h"
#include "ParamsMapping.h"
#include "ParamsDeBruijn.h"
#include "ParameterizedReadFile.h"
#include "ParameterizedInitialFile.h"
#include "ParamsCycleManagement.h"
#include "OutputFileNull.h"
#include "OutputFileFasta.h"
#include "OutputFilePriceq.h"
#include "Assembler.h"
#include "AssemblerListener.h"
#include "AssemblerListenerNull.h"
#include "AssemblerListenerStd.h"
#include "AssemblerListenerVerbose.h"
#include "AssemblerListenerCarrier.h"
#include "EcoFilter.h"
#include "ReadPairFilter.h"

#include <set>
#include <sstream>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <queue>
#include <string.h>
#include <bitset>
#include <cstddef>
#include <omp.h>
using namespace::std;
using ::size_t;

string helpString = string(
"PRICE v0.2\n"
"\n"
"These are the command options for the PRICE assembler. \n"
"For more details, see the accompanying README.txt file. \n"
"Contact: price@derisilab.ucsf.edu \n"
"\n"
"Usage: ./PriceTI [args] \n"
"\n"
"INPUT FILES: \n"
" accepted formats are fasta (.fa or .fasta), fastq (.fq, .fastq, or _sequence.txt), or priceq (.pq or .priceq) \n"
"INPUT READ FILES: \n"
"  NOTE: these flags can be used multiple times in the same command to include multiple read datasets. \n"
"  (default % ID = 90%) \n"
" -fp a b c [d e]: (a,b)input file pair, (c)amplicon insert size (including read) \n"
"                  (d,e) are optional; (d)the initial cycle at which this file will be used, (e)the num cycles for which it is used \n"
" -fpp a b c d [e f]: (a,b)input file pair, (c)amplicon insert size (including read), (d)required % identity for match (25-100 allowed) \n"
"                     (e,f) are optional; (e)the initial cycle at which this file will be used, (f)the num cycles for which it is used \n"
"INPUT INITIAL CONTIG FILES: \n"
"  NOTE: these flags can be used multiple times in the same command to include multiple initial contig datasets. \n"
" -icf a b c d: (a)initial contig file, (b)number of addition steps, (c)number of cycles per step, (d)const by which to multiply quality scores \n"
" -picf a b c d e: (a)num of initial contigs from this file, (b)initial contig file, (c)num addition steps, (d)num cycles per step (e)const by which to multiply quality scores \n"
"OUTPUT FILES: \n"
" accepted formats are fasta (.fa or .fasta) or priceq (.pq or .priceq) \n"
" -o a: (a)output file name (.fasta or .priceq) \n"
" -nco a: (a)num. cycles that pass in between output files being written (default=1) \n"
"OTHER PARAMS: \n"
" -nc a: (a)num. of cycles \n"
" -link a: (a)max. number of contigs that are allowed to replace a read in a mini-assembly (default=5)\n"
" -mol a: (a)minimum overlap length for mini-assembly (default=35) \n"
" -tol a: (a)threshold seq num for scaling overlap for overhang assemblies (default=20)\n"
" NOTE: -mol and -tol do not affect the parameters for de-Bruijn-graph-based assembly.\n"
" -mpi a: (a)minimum % identity for mini-assembly (default=85) \n"
" -tpi a: (a)threshold seq num for scaling % ID for mini-assemblies (default=20)\n"
" -MPI, -TPI : same as above, but for meta-assembly (-MPI default=85, -TPI default=1000)\n"
" NOTE: there is no minimum overlap value for meta-assembly\n"
" -trim a b: after (a) cycles, (b)min. coverage level, (optional min. length after trim)\n"
"            if (a) == 0, initial contigs will be trimmed whenever they are added\n"
" -target a b [c d]: limit output contigs to those with matches to input initial contigs at the end of each cycle.\n"
"                    (a) % identity to an input initial contig to count as a match (ungapped); (b)num cycles to skip\n"
"                    before applying this filter.  [c and d are optional, but must both be provided if either is]\n"
"                    After target filtering has begin, target-filtered/-unfiltered cycles will alternate with (c)\n"
"                    filtered cycles followed by (d) unfiltered cycles.\n"
" -targetF a b [c d]: the same as -target, but now matches to all reads in the input set will be specified, not just\n"
"                    the ones that have been introduced up to that point (this is FullFile mode).\n"
" -dbmax a: (a) the maximum length sequence that will be fed into de Bruijn assembly \n"
"           (default=100; recommended: max paired-end read length)\n"
" -dbk a: (a) the k-mer size for de Bruijn assembly (default=20; keep less than the read length)\n"
" -dbms a: (a) the minimum number of sequences to which de Bruijn assembly will be applied (default=3)\n"
" -lenf a b: filters out contigs shorter than (a) nt at the end of every cycle, after skipping (b) cycles.\n"
" -maxHp a: filters out a pair of reads if either read has a homo-polymer track >(a) nucleotides in length.\n"
" -reset a [b c d...]: re-introduces contigs that were previously not generating assembly jobs of their own\n"
"                      (a) is the one-indexed cycle where the contigs will be reset.  Same with b, c, d. \n"
"                      Any number of args may be added.\n"
"COMPUTATIONAL EFFICIENCY: \n"
" -a x: (x)num threads to use (default=1) \n"
" -mtpf a: (a)max threads per file (default=1) \n"
"USER INTERFACE: \n"
" -log a: determines the type of outputmakes the output verbose (lots of time stamp tags) \n"
"         (a) = c: concise stdout (default)\n"
"         (a) = n: no stdout \n"
"         (a) = v: verbose stdout \n"
" -logf a: (a)the name of an output file for verbose log info to be written (doesn't change stdout format) \n"
" -, -h, or --help: user interface info. \n"
);

// returns a string with the file type based on the append of the filename
// NONEFILE: no described filetype
// FASTAFILE: .fa or .fasta
// FASTQFILE: .fq or .fastq or _sequence.txt
enum FileType { NONEFILE, FASTAFILE, FASTQFILE, PRICEQFILE };
FileType getFileType(string filename){
  size_t lastDot = filename.rfind('.');
  if (lastDot == string::npos){ return NONEFILE; }
  else {
    string append = filename.substr(lastDot);
    if (append == ".fa" or append == ".fasta"){ return FASTAFILE; }
    else if (append == ".fq" or append == ".fastq"){ return FASTQFILE; }
    else if (append == ".pq" or append == ".priceq"){ return PRICEQFILE; }
    else if (append == ".txt"){
      size_t underscore = filename.rfind('_');
      if (underscore == string::npos){ return NONEFILE; }
      else {
	string uAppend = filename.substr(underscore);
	if (uAppend == "_sequence.txt"){ return FASTQFILE; }
	else { return NONEFILE; }
      }
    }
    else { return NONEFILE; }
  }
}

int main(int argc, char *argv[]){

  set<ParameterizedReadFile*> readFiles;
  set<ParameterizedInitialFile*> initContigFiles;
  int initialInput = 0;
  bool willDoAssembly = true; // allows full set of errors to be printed
  OutputFile* outfile = new OutputFileNull();
  bool distributeInitialContigs = false;
  int cyclesPerStep;
  int numSteps;

  // default is 1 core
  omp_set_num_threads( 1 );
  int maxThreadsPerFile = 1;

  // these may be changed
  ParamsMinOverlap* pmoDefault = Assembler::getDefaultParamsMinOverlap();
  ParamsMinFractId* pmfDefault = ParamsMinFractId::getDefault();

  ParamsMinOverlap* pmoMini = Assembler::getDefaultParamsMinOverlap();
  ParamsMinFractId* pmfMini = ParamsMinFractId::getDefault();
  AlignmentScoreMatrix* asmMini = AlignmentScoreMatrix::getDefault();

  ParamsMinOverlap* pmoMeta = new ParamsMinOverlap(200,
						   pmoDefault->minOvlThreshold(),
						   pmoDefault->minOvlSlope());
  
  ParamsMinFractId* pmfMeta = new ParamsMinFractId(pmfDefault->globalFractId(),
						   1000,
						   pmfDefault->denominatorLogStep());
  AlignmentScoreMatrix* asmMeta = AlignmentScoreMatrix::getDefault();
  delete pmoDefault;
  delete pmfDefault;

  int deBruijnKmerSize = ParamsDeBruijn::defaultKmerSize();
  long deBruijnMaxSeqLength = ParamsDeBruijn::defaultMaxSeqLength();
  long deBruijnMinSeqNum = ParamsDeBruijn::defaultMinNumSeqs();

  // this first one doesn't have a command yet, and maybe shouldn't
  EcoFilterInitialContigMatch* ecoFilterTarget = 0;
  EcoFilterMinLength* ecoFilterLength = 0;

  // starts out as null, may be replaced (must at least be replaced by an RpfNull object
  ReadPairFilter* rpfHomoPolymer = NULL;

  AssemblerListener* listenerUser = new AssemblerListenerStd();
  AssemblerListener* listenerFile = new AssemblerListenerNull();


  // first, get the total number of cycles and create the cycle manager
  int numOfCyclesToDo = 0;
  long maxNumLinks = ParamsCycleManagement::defaultMaxContigLinkages();
  int argn = 1;
  while ( argn < argc ){
    string arg = string(argv[argn]);
    if (arg == "-nc"){
      numOfCyclesToDo = atoi( argv[argn+1] );
      argn += 2;
    } else if (arg == "-link"){
      maxNumLinks = atol( argv[argn+1] );
      argn += 2;
    } else { argn++; }
  }
  ParamsCycleManagement* cycleInfo = new ParamsCycleManagement(numOfCyclesToDo,maxNumLinks);

  // now go through again and get the command-line args organized
  argn = 1;
  while ( argn < argc ){
    string arg = string(argv[argn]);
    if (arg == "-fp" or arg == "-fpp"){
      string fileA = argv[argn+1];
      FileType fileTypeA = getFileType(fileA);
      ReadFile* rfHalfA;
      if (fileTypeA == FASTQFILE){ rfHalfA = new ReadFileFastqSingle(fileA); }
      else if (fileTypeA == PRICEQFILE){ rfHalfA = new ReadFilePriceqSingle(fileA); }
      else if (fileTypeA == FASTAFILE){ rfHalfA = new ReadFileFastaSingle(fileA); }
      else { throw AssemblyException::ArgError("read file is of an inappropriate type."); }

      string fileB = argv[argn+2];
      FileType fileTypeB = getFileType(fileB);
      ReadFile* rfHalfB;
      if (fileTypeB == FASTQFILE){ rfHalfB = new ReadFileFastqSingle(fileB); }
      else if (fileTypeB == PRICEQFILE){ rfHalfB = new ReadFilePriceqSingle(fileB); }
      else if (fileTypeB == FASTAFILE){ rfHalfB = new ReadFileFastaSingle(fileB); }
      else { throw AssemblyException::ArgError("read file is of an inappropriate type."); }

      long ampSize = atoi(argv[argn+3]);
      ParamsMapping* mapParams;

      // helps deal with the optional last two args
      int lastArgN;
      if (arg == "-fp"){
	lastArgN = 3;
	mapParams = ParamsMapping::getDefault();
      } else if (arg == "-fpp"){
	lastArgN = 4;
	int pctId = atoi(argv[argn+4]);
	if (pctId < 25 or pctId > 100){
	  throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100");
	}
	ParamsMapping* defaultMapParams = ParamsMapping::getDefault();
	mapParams  = new ParamsMapping(defaultMapParams->stepSize(), float(pctId) / 100.0 );
	delete defaultMapParams;
      }
      else { throw AssemblyException::LogicError("missing param option for -fp or -fpp"); }

      ReadFile* rf = new ReadFileDual(rfHalfA, rfHalfB, ampSize);

      int advanceArgN;
      if ( argn + lastArgN + 1 >= argc or argv[argn + lastArgN + 1][0] == '-'){
	readFiles.insert( new ParameterizedReadFile( rf, mapParams ) );
	advanceArgN = lastArgN + 1;
      } else {
	int startCycle = atoi(argv[argn+lastArgN+1]);
	int numCycles = atoi(argv[argn+lastArgN+2]);
	readFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, numCycles ) );
	advanceArgN = lastArgN + 3;
      }
      delete rf;
      delete rfHalfA;
      delete rfHalfB;
      delete mapParams;
      argn += advanceArgN;
   } else if (arg == "-icf"){
      string filename = argv[argn+1];
      FileType filetype = getFileType(filename);
      float countFactor = atof( argv[argn+4] );
      ReadFile* rf;
      if (filetype == FASTQFILE){ rf = new ReadFileFastqSingle(filename,countFactor); }
      else if (filetype == PRICEQFILE){ rf = new ReadFilePriceqSingle(filename); }
      else if (filetype == FASTAFILE){ rf = new ReadFileFastaSingle(filename,countFactor); }
      else { throw AssemblyException::ArgError("from -icf: read file is of an inappropriate type."); }
      int numSteps = atoi( argv[argn+2] );
      int cyclesPerStep = atoi( argv[argn+3] );
      initContigFiles.insert( new ParameterizedInitialFile( rf, numSteps, cyclesPerStep ) );
      delete rf;
      argn += 5;
    } else if (arg == "-picf"){
      long initialInput = atol( argv[argn+1] );
      string filename = argv[argn+2];
      FileType filetype = getFileType(filename);
      float countFactor = atof( argv[argn+5] );
      ReadFile* rf;
      if (filetype == FASTQFILE){ rf = new ReadFileFastqSingle(filename,countFactor); }
      else if (filetype == PRICEQFILE){ rf = new ReadFilePriceqSingle(filename); }
      else if (filetype == FASTAFILE){ rf = new ReadFileFastaSingle(filename,countFactor); }
      else { throw AssemblyException::ArgError("from -picf: read file is of an inappropriate type."); }
      int numSteps = atoi( argv[argn+3] );
      int cyclesPerStep = atoi( argv[argn+4] );
      initContigFiles.insert( new ParameterizedInitialFile( initialInput, rf, numSteps, cyclesPerStep ) );
      delete rf;
      argn += 6;
    } else if (arg == "-nc"){
      // nothing happens here because this info was already collected
      //numOfCyclesToDo = atoi( argv[argn+1] );
      argn += 2;
    } else if (arg == "-target" or arg == "-targetF"){
      bool fullFileMode = (arg == "-targetF");
      float fractId = atof( argv[argn+1] ) / float(100.0);
      int cyclesToSkip = atoi( argv[argn+2] );
      if ( argn + 3 >= argc or argv[argn+3][0] == '-'){
	delete ecoFilterTarget;
	ecoFilterTarget = new EcoFilterInitialContigMatch(fractId,cyclesToSkip,fullFileMode);
	argn += 3;
      } else {
	int numFilterCycles = atoi( argv[argn+3] );
	int numSkipCycles = atoi( argv[argn+4] );
	delete ecoFilterTarget;
	ecoFilterTarget = new EcoFilterInitialContigMatch(fractId,cyclesToSkip,numFilterCycles,numSkipCycles,fullFileMode);
	argn += 5;
      }
    } else if (arg == "-maxHp"){
      delete rpfHomoPolymer;
      rpfHomoPolymer = new ReadPairFilterHomoPolymer( atol(argv[argn+1]) );
      argn += 2;
    } else if (arg == "-lenf"){
      long minLength = atol( argv[argn+1] );
      int cyclesToSkip = atoi( argv[argn+2] );
      delete ecoFilterLength;
      ecoFilterLength = new EcoFilterMinLength(minLength,cyclesToSkip);
      argn += 3;
    } else if (arg == "-link"){
      // nothing happens; already dealt with
      argn += 2;
    } else if (arg == "-o"){
      delete outfile;
      string filename = argv[argn+1];
      FileType fileType = getFileType(filename);
      if (fileType == PRICEQFILE){
	outfile = new OutputFilePriceq( filename );
      } else if (fileType == FASTAFILE){
	outfile = new OutputFileFasta( filename );
      } else {
	throw AssemblyException::ArgError("output filenames must be fasta or priceq");
      }
      argn += 2; 
    } else if (arg == "-nco"){
      cycleInfo->setOutputInterval( atoi( argv[argn+1] ) );
      argn += 2; 
    } else if (arg == "-a"){
      omp_set_num_threads( atoi( argv[argn+1] ) );
      argn += 2; 
    } else if (arg == "-dbk"){
      deBruijnKmerSize = atoi( argv[argn+1] );
      argn += 2; 
    } else if (arg == "-dbmax"){
      deBruijnMaxSeqLength = atol( argv[argn+1] );
      argn += 2; 
    } else if (arg == "-dbms"){
      deBruijnMinSeqNum = atol( argv[argn+1] );
      argn += 2; 
    } else if (arg == "-mtpf"){
      maxThreadsPerFile = atoi( argv[argn+1] );
      argn += 2; 
    } else if (arg == "-trim"){
      int numCycles = atoi( argv[argn+1] );
      float minCoverage = atof( argv[argn+2] );
      if ( argn + 3 < argc and argv[argn+3][0] != '-' ) {
	long minLength = atol( argv[argn+3] );
	cycleInfo->addTrim(numCycles, minCoverage, minLength);
	argn += 4;
      } else {
	cycleInfo->addTrim(numCycles, minCoverage);
	argn += 3;
      }
    } else if (arg == "-reset"){
      int numResets = 0;
      argn += 1;
      while (argn < argc and argv[argn][0] != '-'){
	cycleInfo->addContigReset( atoi( argv[argn] ) );
	argn += 1;
	numResets++;
      }
      if (numResets == 0 ){
	throw AssemblyException::ArgError("-reset should have at least one cycle number after it.");
      }
    } else if (arg == "-mol"){ // min overlap for overhang assemblies
      long minOvl = atol( argv[argn+1] );
      ParamsMinOverlap* pmoMiniHolder = new ParamsMinOverlap(minOvl,
							     pmoMini->minOvlThreshold(),
							     pmoMini->minOvlSlope());
      delete pmoMini;
      pmoMini = pmoMiniHolder;
      argn += 2; 
    } else if (arg == "-tol"){ // threshold for scaling overlap for overhang assemblies
      long ovlThreshold = atol( argv[argn+1] );
      ParamsMinOverlap* pmoMiniHolder = new ParamsMinOverlap(pmoMini->globalMinOverlap(),
							     ovlThreshold,
							     pmoMini->minOvlSlope());
      delete pmoMini;
      pmoMini = pmoMiniHolder;
      argn += 2; 
    } else if (arg == "-mpi"){ // min percent ID for overhang assemblies
      int pctId = atoi(argv[argn+1]);
      if (pctId < 25 or pctId > 100){
	throw AssemblyException::ArgError("% ID required for mini-assembly is out of range 25-100");
      }
      ParamsMinFractId* pmfMiniHolder = new ParamsMinFractId(float(pctId) / 100.0,
							     pmfMini->fractIdThreshold(),
							     pmfMini->denominatorLogStep());
      delete pmfMini;
      pmfMini = pmfMiniHolder;
      argn += 2; 
    } else if (arg == "-tpi"){ // threshold for scaling percent ID for overhang assemblies
      long threshold = atol(argv[argn+1]);
      ParamsMinFractId* pmfMiniHolder = new ParamsMinFractId(pmfMini->globalFractId(),
							     threshold,
							     pmfMini->denominatorLogStep());
      delete pmfMini;
      pmfMini = pmfMiniHolder;
      argn += 2; 
    } else if (arg == "-MPI"){ // min percent ID for meta assemblies
      int pctId = atoi(argv[argn+1]);
      if (pctId < 25 or pctId > 100){
	throw AssemblyException::ArgError("% ID required for mini-assembly is out of range 25-100");
      }
      ParamsMinFractId* pmfMetaHolder = new ParamsMinFractId(float(pctId) / 100.0,
							     pmfMeta->fractIdThreshold(),
							     pmfMeta->denominatorLogStep());
      delete pmfMeta;
      pmfMeta = pmfMetaHolder;
      argn += 2; 
    } else if (arg == "-TPI"){ // threshold for scaling percent ID for meta assemblies
      long threshold = atol(argv[argn+1]);
      ParamsMinFractId* pmfMetaHolder = new ParamsMinFractId(pmfMeta->globalFractId(),
							     threshold,
							     pmfMeta->denominatorLogStep());
      delete pmfMeta;
      pmfMeta = pmfMetaHolder;
      argn += 2; 
   } else if (arg == "-log"){
      delete listenerUser;
      string type = string(argv[argn+1]);
      if (type == "n"){ listenerUser = new AssemblerListenerNull(); }
      else if (type == "c"){ listenerUser = new AssemblerListenerStd(); }
      else if (type == "v"){ listenerUser = new AssemblerListenerVerbose(); }
      else { cout << "'" << type << "' is not a valid stdout type ('n', 'c', or 'v')." << endl; }
      argn += 2;
   } else if (arg == "-logf"){
      delete listenerFile;
      listenerFile = new AssemblerListenerVerbose( argv[argn+1] );
      argn += 2;
   } else if (arg == "-h" or arg == "-" or arg == "--help"){
      willDoAssembly = false;
      cout << helpString << endl;
      argn++;
    } else {
      cout << "'" << arg << "' is an invalid argument." << endl;
      throw AssemblyException::ArgError("died due to bad arg flag (see stdout above).");
      argn++;
    }
    if (argn != argc and argv[argn][0] != '-'){
      ostringstream outMessage;
      outMessage << arg << " was given an incorrect number of args.";
      throw AssemblyException::ArgError( outMessage.str().c_str() );
    }
  }

  // add the filters to the cycle controller
  EcoFilter* defaultEcoFilterMinCount = EcoFilterMinCount::defaultFilter();
  cycleInfo->addEcoFilter(defaultEcoFilterMinCount);
  delete defaultEcoFilterMinCount;
  if (ecoFilterLength != NULL){
    cycleInfo->setSizeFilter(ecoFilterLength);
    delete ecoFilterLength;
  }
  if (ecoFilterTarget != NULL){
    ecoFilterTarget->addFiles(&initContigFiles);
    cycleInfo->addEcoFilter(ecoFilterTarget);
    delete ecoFilterTarget;
  }

  ReadPairFilter* rpf;
  if (rpfHomoPolymer == NULL){ rpf = new ReadPairFilterNull(); }
  else { rpf = rpfHomoPolymer; }

  AssemblerListener* listenerArray[] = {listenerUser, listenerFile};
  AssemblerListener* omniListener = new AssemblerListenerCarrier(listenerArray, 2);

  // TEMPORARY set-up for returning the command that launched the job
  ostringstream launchCommand;
  launchCommand << "Launch command:";
  for (int n = 0; n < argc; ++n){ launchCommand << " " << argv[n]; }
  omniListener->message( launchCommand.str().c_str() );

  if ( willDoAssembly ){

    // do checks to make sure all of the initial values are legit???

    ParamsAlignment* miniAlignParams = new ParamsAlignment(pmoMini, pmfMini, asmMini);
    // NOTE: pmoMeta is just a filler; this class should be able to take a NULL for
    // that variable if no min overlap will be used
    ParamsAlignment* metaAlignParams = new ParamsAlignment(pmoMeta, pmfMeta, asmMeta);
    ParamsDeBruijn* deBruijnParams = new ParamsDeBruijn(deBruijnKmerSize, ParamsDeBruijn::defaultMinCount(), deBruijnMaxSeqLength, deBruijnMinSeqNum);

    Assembler* assembler = new Assembler(&readFiles, &initContigFiles, cycleInfo, outfile, omniListener,
					 maxThreadsPerFile, miniAlignParams, metaAlignParams, deBruijnParams, rpf);
    delete miniAlignParams;
    delete metaAlignParams;
    delete deBruijnParams;
    assembler->runAssembly();

    set<ScoredSeq*> gotten;
    assembler->getContigs( &gotten );
    ostringstream numFinalContigs;
    numFinalContigs << "Num final contigs: " << gotten.size();
    omniListener->message( numFinalContigs.str().c_str() );
    for (set<ScoredSeq*>::iterator it = gotten.begin(); it != gotten.end(); ++it){
      (*it)->deepDelete();
    }

    delete assembler;
    omniListener->message("THE END");
  }
  delete cycleInfo;
  delete pmoMini;
  delete pmfMini;
  delete asmMini;
  delete pmoMeta;
  delete pmfMeta;
  delete asmMeta;
  delete outfile;
  delete omniListener;
  delete rpf;
  for (set<ParameterizedReadFile*>::iterator it = readFiles.begin(); it != readFiles.end(); ++it){
    delete *it;
  }
  for (set<ParameterizedInitialFile*>::iterator it = initContigFiles.begin(); it != initContigFiles.end(); ++it){
    delete *it;
  }
  return 0;
}



